/* This is the declaration of a class to manage plugins for VideoInputDevice's
   classes.

   There are three main groups of functions:
   * those to load plugins;
   * those to query the manager for the list of devices it supports;
   * those to get a PVideoInputDevice* to an already initialized object;

   The interface to get a PVideoInputDevice will show the dichotomy between
   two types of plugins:
   * those that come with pwlib, upon which the client program should know
   little, and fully rely on the manager to get them;
   * those that come with the client program, that the client program may want
   to have a more direct access to;

   The exact details of how to code a plugin depend on the implementation of
   that class; see src/ptlib/unix/vidinput_manager_dlopen.cxx for example.

   There are two types of names to get to a real device:
   * full name "DRIVER SHORTNAME"; the manager strips the driver name and gives
   the short name to the corresponding driver; that allows a single SHORTNAME
   to be handled by several drivers;
   * short name, for example "/dev/video3.14" or "PiCam"; this is the name that
   only plugins are supposed to understand; as such, the manager should try
   to scan the plugins for one that accepts it, and use the first that does.

*/

class PVideoInputManager
{
public:
  /* those functions should never get called: the typical use of the manager
     is through a direct call to the static member functions */
  PVideoInputManager();
  ~PVideoInputManager();

  
  /* These functions are there to load plugins; either by pointing the manager
     to a directory where they are to be found, or directly to the filename.

     LoadPlugin returns TRUE if the plugin could be loaded and FALSE otherwise
     LoadPluginDirectory returns FALSE only if no plugin could be loaded
   */
  static BOOL LoadPluginDirectory(PDirectory directory_name);
  static BOOL LoadPlugin(PString plugin_filename);

  /* This single function will return the list of all supported devices,
     in a correctly mangled way; for example "V4L /dev/video3.14"; the first
     part is the plugin name "V4L" that handles the device, and
     "/dev/video3.14" is either directly a filename, or a name that the plugin
     will know how to use anyway "PiCam", that the plugin will relate to
     "/dev/video3.14" (this is a 'userfriendly names' feature; the V4L and AVC
     plugin already come with it).
  */
  static PStringList GetInputDeviceNames();

  /* Those two functions are the ones used to get a pointer to an usual,
     already created PVideoInputDevice (ie: equivalent to a
     'new PVideoInputDevice').

     GetDevice should be used in case the program already asked for a plugin
     through LoadPlugin, and wants direct access to it; it returns a pointer to
     a non-opened PVideoInputDevice, since it is assumed the program knows the
     plugin, and how to directly talk with it;

     GetOpenedDevice should be used in case the program has no particular
     knowledge about the plugin; as such it takes the full device name (ie:
     "V4L PiCam"), and the BOOL startImmediate of the usual call to
     PVideoInputDevice::Open, then return a pointer to an ALREADY OPENED 
     PVideoInputDevice.
   */  
  static PVideoInputDevice *GetDevice(PString plugin_name);
  static PVideoInputDevice *GetOpenedDevice(PString full_name,
					    BOOL startImmediate);

};

#define DECLARE_VIDEOINPUT_PLUGIN(NAME, CLASS) \
extern "C" char *plugin_type(void) { return "VideoInput"; }; \
extern "C" char *plugin_getname(void) { return NAME; }; \
extern "C" PVideoInputDevice *plugin_create() { return new CLASS(); } \
extern "C" PStringList plugin_getdevices() { return CLASS::GetInputDeviceNames(); }

