/* This is an implementation of the PVideoInputManager class,
   that should work on any OS providing the dlopen&dlsym calls.

   It stores the list in a custom one-linked list of plugins, that
   is almost read-only: only one function writes to it, LoadPlugin,
   and it only adds plugins at the beginning of the list.

   Once a plugin is loaded, it is for the life of the program: the list
   is never, ever, explicitly freed.

   A plugin for this manager comes in the form of (those are AND):
   * shared lib;
   * the implementation of a class derived from PVideoInputDevice;
   * some extern "C"-defined functions;

   To get more specific, a typical plugin comes a a .h, declaring a class as
   public PVideoInputDevice, and a .cxx, that defines the class, and the
   following functions:

   extern "C" char *plugin_type(void) { return "VideoInput"; };
 
   extern "C" char *plugin_getname(void) { return "PI"; };
 
   extern "C" PVideoInputDevice *plugin_create()
   {
     return new PVideoInputPiDevice();
   }
 
   extern "C" PStringList plugin_getdevices()
   {
     return PVideoInputPiDevice::GetInputDeviceNames();
   }

   where:
   * plugin_type helps the manager being sure it loads the right type of
   plugins;
   * plugin_getname allows the manager to demangle the full names;
   * plugin_create is used to get an instance of the plugin's class;
   * plugin_getdevices returns the list of supported devices (short names,
   the manager will take care of adding the plugin name in front of it);

*/

// system includes
#include <dlfcn.h>

// PWLib includes
#include <ptlib.h>
#include <ptlib/videoio.h>
#include <ptlib/vconvert.h>
#include <ptlib/pdirect.h>

struct video_plugin {
  const char *name;
  void *handle;
  PVideoInputDevice *(*Create)(void);
  PStringList (*DeviceNames)(void);
  struct video_plugin *next;
};

static void LoadAllPlugins(); // helper function
static struct video_plugin *plugins_list;  
static PMutex mutex;

PVideoInputManager::PVideoInputManager()
{
  PTRACE(3, "PVideoInputManager::PVideoInputManager called: useless...");
  mutex.Wait();
  if(plugins_list == NULL)
    {
      mutex.Signal();
      LoadAllPlugins();
    }
  else
    mutex.Signal();
}


PVideoInputManager::~PVideoInputManager()
{
  // well, you're not really supposed to create one anyway...
}

void
LoadAllPlugins()
{
  PTRACE(3, "Loading plugins...");
  if(PVideoInputManager::LoadPluginDirectory("/usr/lib/pwlib/video") == FALSE)
    {
      PTRACE(2, "Couldn't load any system plugin!");
    }
}

BOOL
PVideoInputManager::LoadPluginDirectory(PDirectory plugdir)
{
  BOOL Result = FALSE;

  PTRACE(4, "Considering loading plugins in " << plugdir);
  
  if(!plugdir.Open())
    {
      PTRACE(4, "the dir isn't accessible");
      return Result; // no love, no hope
    }
  
  do
    {
      PString entry = plugdir+plugdir.GetEntryName();
      if(plugdir.IsSubDir())
	Result = LoadPluginDirectory(entry) || Result; // BEWARE of the...
      else
	Result = LoadPlugin(entry) || Result; // lazy evaluation!!!
    } 
  while(plugdir.Next());

  return Result;
  
}

BOOL
PVideoInputManager::LoadPlugin(PString filename)
{
      void *handle = NULL;
      struct video_plugin *plugin = NULL;

      ostream & s = PTrace::Begin(0, "Plugin manager: ", 0);

      plugin = (struct video_plugin *)malloc(sizeof(struct video_plugin));
      if(plugin == NULL)
	{
	  s << "PVideoInputManager::LoadPlugin: ENOMEM\n" << PTrace::End(s);
	  return FALSE;
	}

      s << "trying to load " << filename << "\n";

      // first, we try to get a handle on the plugin
      handle=dlopen(filename, RTLD_NOW);
      if(handle == NULL)
	{
	  s << "dlerror=" << dlerror() << "\n";
	  free(plugin);
	  s << PTrace::End(s);
	  return FALSE;
	}
      plugin->handle = handle;

      // second, we check its type
      char *(*gettype)(void);
      gettype = (char *(*)())dlsym(handle, "plugin_type");
      if(gettype == NULL)
	{
	  s << "dlerror=" << dlerror() << "\n";
	  goto wrong;
	}
      if(strcmp("VideoInput", (*gettype)()) != 0)
	{
	  s << "not a VideoInput plugin\n";
	  goto wrong; // not exactly wrong, but at least no what was sought
	}

     // third, we try to get its name
      char *(*getname)(void);
      getname = (char *(*)())dlsym(handle, "plugin_getname");
      if(getname == NULL)
	{
	  s << "dlerror=" << dlerror() << "\n";
	  goto wrong;
	}
      plugin->name = (*getname)();
	
      // fourth, we try to get its creator
      PVideoInputDevice *(*Create)();
      Create = (PVideoInputDevice *(*)(void))dlsym(handle,"plugin_create");
      if(Create == NULL)
	{
	  s << "dlerror=" << dlerror() << "\n";
	  goto wrong;
	}
      plugin->Create = Create;

      // fourth, we try to get its devices lister
      PStringList (*DeviceNames)();
      DeviceNames = (PStringList (*)(void))dlsym(handle, "plugin_getdevices");
      if(DeviceNames == NULL)
	{
	  s << "dlerror=" << dlerror() << "\n";
	  goto wrong;
	}
      plugin->DeviceNames = DeviceNames;
   
      // finally, since we got all we wanted, we just add it to the list!
      mutex.Wait();
      plugin->next = plugins_list;
      plugins_list = plugin;
      mutex.Signal();
 
      s << "success\n";
     
      return TRUE;
 wrong: 
      free(plugin);
      s << PTrace::End(s);
      (void)dlclose(handle);
      return FALSE;
}

PVideoInputDevice*
PVideoInputManager::GetDevice(PString name)
{ 
  mutex.Wait();
  struct video_plugin *ptr=plugins_list;
  mutex.Signal();
  for(;
      ptr != NULL;
      ptr = ptr->next)
    {
      if(strcmp(name, ptr->name) == 0)
	return ptr->Create();
    }

  return NULL;
}

PVideoInputDevice*
PVideoInputManager::GetOpenedDevice(PString devName, BOOL flag)
{
  char driver[11]; // FIXME! Only ten chars for the driver name...

  mutex.Wait();
  if(plugins_list == NULL)
    {
      mutex.Signal();
      LoadAllPlugins();
    }
  else
    mutex.Signal();


  if(sscanf(devName, "%s", driver) == 1)
    { // looks like a nice devName
      PString filename = devName.Right(strlen(devName)-strlen(driver)-1);
      // FIXME! this computation is awful!

      PVideoInputDevice *dev = GetDevice(driver);

      if(dev != NULL && dev->Open(filename, flag))
	return dev;

      if(dev != NULL)
	delete dev;

      return NULL;
    }
  else
    { // doesn't look like a nice devName
      // (that can be a user on CLI, who wants us to open /dev/video3.14)

      // FIXME: there is a nicer solution to implement that,
      // but my pattern-matching abilities are quite small,
      // as the previous fixme must have proven: search through the list of
      // devices names we get from GetInputDeviceNames, and only ask to those
      // drivers that claimed they could handle devName.

      mutex.Wait();
      struct video_plugin *ptr=plugins_list;
      mutex.Signal();
      for(;
	  ptr != NULL;
	  ptr = ptr->next)
	{
	  PVideoInputDevice *dev=ptr->Create();
	  if(dev != NULL && dev->Open(devName, flag))
	    return dev;

	  if(dev != NULL)
	    delete dev;
	}

      return NULL;
    }
}

PStringList
PVideoInputManager::GetInputDeviceNames()
{
  PTRACE(3, "PVideoInputManager::GetInputDeviceNames called");
 
  mutex.Wait();
  if(plugins_list == NULL)
    {
      mutex.Signal();
      LoadAllPlugins();
    }
  else
    mutex.Signal();

  // we get through all the plugin lists, and add the name as a prefix,
  // to be able to track down which takes care of what
  
  PStringList Result;
 
  mutex.Wait();
  struct video_plugin *ptr = plugins_list;
  mutex.Signal();
  for(;
      ptr != NULL;
      ptr=ptr->next)
    {
      PStringList tmplist = ptr->DeviceNames();

      for (PINDEX i = 0; i < tmplist.GetSize(); i++)
	{
	  Result.AppendString((PString)ptr->name+" "+tmplist[i]);
	}
    }
 
  return Result;

}
